/************************************************************************/
/*                                                                      */
/*    *****                       *****                                 */
/*      *****                   *****                                   */
/*        *****               *****                                     */
/*          *****           *****                                       */
/*  ***************       ***************                               */
/*  *****************   *****************                               */
/*  ***************       ***************                               */
/*          *****           *****           TheNetNode                  */
/*        *****               *****         Portable                    */
/*      *****                   *****       Network                     */
/*    *****                       *****     Software                    */
/*                                                                      */
/* File os/linux/ax25ip.c (maintained by: DG9OBU)                       */
/*                                                                      */
/* This file is part of "TheNetNode" - Software Package                 */
/*                                                                      */
/* Copyright (C) 1998 - 2004 NORD><LINK e.V. Braunschweig               */
/*                                                                      */
/* This program is free software; you can redistribute it and/or modify */
/* it under the terms of the NORD><LINK ALAS (Allgemeine Lizenz fuer    */
/* Amateurfunk Software) as published by Hans Georg Giese (DF2AU)       */
/* on 13/Oct/1992; either version 1, or (at your option) any later      */
/* version.                                                             */
/*                                                                      */
/* This program is distributed WITHOUT ANY WARRANTY only for further    */
/* development and learning purposes. See the ALAS (Allgemeine Lizenz   */
/* fuer Amateurfunk Software).                                          */
/*                                                                      */
/* You should have received a copy of the NORD><LINK ALAS (Allgemeine   */
/* Lizenz fuer Amateurfunk Software) along with this program; if not,   */
/* write to NORD><LINK e.V., Hinter dem Berge 5, D-38108 Braunschweig   */
/*                                                                      */
/* Dieses Programm ist PUBLIC DOMAIN, mit den Einschraenkungen durch    */
/* die ALAS (Allgemeine Lizenz fuer Amateurfunk Software), entweder     */
/* Version 1, veroeffentlicht von Hans Georg Giese (DF2AU),             */
/* am 13.Oct.1992, oder (wenn gewuenscht) jede spaetere Version.        */
/*                                                                      */
/* Dieses Programm wird unter Haftungsausschluss vertrieben, aus-       */
/* schliesslich fuer Weiterentwicklungs- und Lehrzwecke. Naeheres       */
/* koennen Sie der ALAS (Allgemeine Lizenz fuer Amateurfunk Software)   */
/* entnehmen.                                                           */
/*                                                                      */
/* Sollte dieser Software keine ALAS (Allgemeine Lizenz fuer Amateur-   */
/* funk Software) beigelegen haben, wenden Sie sich bitte an            */
/* NORD><LINK e.V., Hinter dem Berge 5, D-38108 Braunschweig            */
/*                                                                      */
/************************************************************************/

/************************************************************************/
/*                                                                      */
/* AX25IP.C V1.05 01.9.98 by DG1KWA / Andreas                           */
/*                        extended by DG9OBU / Marc                     */
/************************************************************************/

#include "tnn.h"

#ifdef AX25IP
#include "ax25ip.h"

/*************************************************************************/
/* AX25IP-Hausaufgaben : ausstehende Frames senden                       */
/*************************************************************************/
void ax25ip(void)
{
  /* nur wenn Interface aktiv ist */
  if (!ax25ip_active)
    return;

  /* empfangene Frames abholen */
  ax25ip_recv();

  /* ausstehende Frames senden wenn vorhanden */
  if (kick[ax25ip_port])
    ax25ip_send();
}

/*************************************************************************/
/* Einen Eintrag aus der Routingtabelle ausgeben                         */
/*************************************************************************/
void show_rt_entry(struct route_table_entry rp, MBHEAD *bufpoi)
{
  /* erst mal merken, wo wir sind */
  bufpoi->l4time = bufpoi->mbpc;

  /* Check auf Eintrag fuer die default-Route */
  if (rp.callsign[0] != 1)
    putprintf(bufpoi, "%s", call_to_a(&rp.callsign[0]));
  else
    putstr("default", bufpoi);

  putspa(11, bufpoi);

  /* IP-Adresse ausgeben */
  putprintf(bufpoi, "%d.%d.%d.%d", rp.ip_addr[0], rp.ip_addr[1], rp.ip_addr[2], rp.ip_addr[3]);

  putspa(29, bufpoi);

  /* Mode und ggf. UDP-Port ausgeben wenn UDP-Port gesetzt */
  if (rp.udp_port != 0)
  {
    /* Eine UDP-Route */
    putstr("UDP", bufpoi);
    putspa(35, bufpoi);
    putnum(ntohs(rp.udp_port), bufpoi);
  }
  else
    /* Eine IP-Route */
    putstr("IP", bufpoi);

/* TEST DG9OBU */
#ifdef AX25IP_DYNLEARN
  /* Routen-Timeout anzeigen */
  if (rp.timeout != 0)
  {
    ULONG d = rp.timeout;
    ULONG h, m, s;                  /* Stunden, Minuten, Sekunden */

    s = d % 60L;
    d /= 60L;
    m = d % 60L;
    d /= 60L;
    h = d % 24L;

    putspa(40, bufpoi);
    putprintf(bufpoi, " %2lu:%02lu:%02lu", h, m, s); /* hh:mm:ss */
  }
#endif

  putstr("\r", bufpoi);
}

/* TEST DG9OBU */
/* Anzahl von IP- oder UDP-Routen in der Routingtabelle feststellen */
/* (IP-Routen haben keinen UDP-Port gesetzt) */
/* UDP = FALSE -> IP-Routen zaehlen, UDP = TRUE -> UDP-Routen zaehlen */
/* Default-Route wird beachtet */
unsigned int count_routes(BOOLEAN udp)
{
  register unsigned int i;
  register unsigned int j = 0;

  /* Gesamte Routingtabelle durchgehen */
  for (i = 0; i < route_tbl_top; ++i)
  {
    if (   ((udp == TRUE) && (route_tbl[i].udp_port != 0))    /* UDP-Route */
        || ((udp == FALSE) && (route_tbl[i].udp_port == 0)))  /* IP-Route */
      ++j;
  }

  /* Default-Route */
  if (default_route.callsign[0] != 0)
  {
    if (   ((udp == TRUE) && (default_route.udp_port != 0))    /* UDP-Route */
        || ((udp == FALSE) && (default_route.udp_port == 0)))  /* IP-Route */
      ++j;
  }

  /* Anzahl Routen melden */
  return (j);
}

/* TEST DG9OBU */
/* Timeout fuer gelernte Routen */
#ifdef AX25IP_DYNLEARN
void route_age(void)
{
  register unsigned int i;

  /* Gesamte Routingtabelle durchgehen */
  for (i = 0; i < route_tbl_top; ++i)
  {
    /* Route mit laufendem Timer */
    if (route_tbl[i].timeout != 0)
    {
      /* Timer laeuft jetzt ab */
      if (--route_tbl[i].timeout == 0)
        /* Route loeschen */
        route_del(route_tbl[i].callsign);
    }
  }
}

/* Partielle Frame-Analyse fuer dyn. Eintraege */
BOOLEAN route_canlearn(MBHEAD *fbp)
{
  char *p;                              /* Zeiger im Header             */
  register int n;
  UBYTE ctl;

  rwndmb(fbp);                          /* Frame von vorne              */

  for (p = rxfhdr, n = 1; n <= L2INUM + L2VNUM; ++n)
  {
    if (!getfid(p, fbp))                /* naechstes Call lesen         */
      return (FALSE);

    p += L2IDLEN;
    if (*(p - 1) & L2CEOA)
      break;                            /* Ende des Addressfeldes       */
  }
  *(p - 1) &= ~L2CEOA;
  *p = NUL;

  ctl = getchr(fbp);                    /* Control-Byte extrahieren     */
  ctl &= ~L2CPF;                        /* Poll-Flag loeschen           */

  /* Nur UI, UA und SABM/SABME erzeugen einen dynamischen Eintrag       */
  if (   ctl == L2CUI                   /* UI                           */
      || ctl == L2CUA                   /* UA                           */
      || ctl == L2CSABM                 /* SABM                         */
#ifdef EAX25
      || ctl == L2CSABME                /* SABME (EAX.25)               */
#endif
     )
    return (TRUE);

  /* Alle anderen Kontrollcodes landen hier */
  return (FALSE);
}
#endif

/*************************************************************************/
/* Kommandointerface fuer Operationen auf die AX25IP-Routentabelle       */
/*************************************************************************/

/* Eine Route anhand eines Calls loeschen, liefert TRUE wenn erfolgreich */
BOOLEAN route_del(unsigned char *call)
{
  register int i;
  register int j;

  /* Soll die Default-Route geloescht werden ? */
  if ((call == NULL) || (call[0] == NUL))
  {
    /* Als unbenutzt markieren */
    default_route.callsign[0] = NUL;

    /* Pruefen, ob Sockets noch notwendig */
    ax25ip_check_down();

    return (TRUE);
  }

  /* Uebergebenes Call in der Tabelle suchen */
  for (i = 0; i < route_tbl_top; ++i)
  {
    if (addrmatch(call, route_tbl[i].callsign))
    {
    /* BINGO ! */
    /* Problem: die Tabelle hat keinen Merker, ob ein Eintrag gueltig ist, */
    /*          die Eintraege haengen einfach hintereinander. Deshalb alle */
    /*          Eintraege oberhalb des gefundenen Eintrages einen nach     */
    /*          unten kopieren und den "table-top" einen herabsetzen.      */
      for (j = i; j < route_tbl_top; ++j)
        route_tbl[j] = route_tbl[j+1];

      /* Ein Tabelleneintrag weniger */
      route_tbl_top--;

      /* Pruefen, ob Sockets noch notwendig */
      ax25ip_check_down();

      return(TRUE);
    }
  }

  return(FALSE);
}

#ifndef AXIP_HELP
static const char inv_mode[] = "Invalid Mode\r";
#else
static const char inv_mode[] = "Syntax:\r"
                               "AXipr +|-|ADD|DEL|DELETE call|DEFAULT [host [UDP [udp-port]]]\r"
                               "AXipr MYUDP [udp-port]\r"
                               "AXipr LOG|LOGLEVEL [loglevel]\r"
  #ifdef AX25IP_DYNLEARN
                               "AXipr TIMEOUT [seconds]\r"
  #endif
                               "AXipr Old\r";
#endif
static const char inv_call[] = "Invalid Callsign\r";
static const char inv_ip[]   = "Invalid IP adress\r";
#ifndef AXIP_HELP
static const char inv_par[]  = "Invalid Parameter\r";
#else
static const char inv_par[] = "Syntax:\r"
                              "AXipr D +|- [host [I|U [udp-port]]]\r"
                              "AXipr R +|- call [host [I|U [udp-port]]]\r"
                              "AXipr U udp-port\r"
                              "AXipr P\r";
#endif

#define HNLEN 64
#define OP_NONE    0
#define OP_ADD     1
#define OP_DEL     2
#define OP_MYUDP   3
#define OP_LOG     4
#define OP_TIMEOUT 5
#define BUFLEN 32

void
ccpaxipr(void)
{
  MBHEAD         *mbp;
  UWORD           port = 0;
  register int    i;
  int             tmp_fd = -1;
  unsigned short  new_udp = 0;
  unsigned char   call[L2IDLEN];
  char            hostname[HNLEN + 1];
  struct          hostent *host;

  char            cBuf[BUFLEN + 1];
  unsigned int    uCmd = OP_NONE;
  unsigned int    uMode = NO_MODE;

  /* neuen Parser verwenden ? */
  if (new_parser == TRUE)
  {

  /* Sysop will aendern und noch was in der Zeile da ? */
  if (issyso() && skipsp(&clicnt, &clipoi))
  {
    /* Frischer Buffer */
    memset(cBuf, 0, sizeof(cBuf));

    /* Operation lesen (add, delete) */
    for (i = 0; i < BUFLEN; ++i)
    {
      if ((!clicnt) || (*clipoi == ' '))
        break;
      clicnt--;
      cBuf[i] = toupper(*clipoi++);
    }

    /* Zum alten Parser zurueckschalten */
    if (   (strcmp(cBuf, "OLD") == 0)
        || (cBuf[0] == 'O')
       )
    {
      new_parser = FALSE;
      putmsg("Switched to old axipr-parser !!!\r");
      return;
    }

    /* Hinzufuegen (add oder +) */
    if (   (strcmp(cBuf, "ADD") == 0)
        || (cBuf[0] == '+')
       )
      uCmd = OP_ADD;

    /* Loeschen (delete, del oder -) */
    if (   (strcmp(cBuf, "DELETE") == 0)
        || (strcmp(cBuf, "DEL") == 0)
        || (cBuf[0] == '-')
       )
      uCmd = OP_DEL;

    /* eigenen UDP-Port aendern */
    if (strcmp(cBuf, "MYUDP") == 0)
      uCmd = OP_MYUDP;

    /* Loglevel */
    if (   (strcmp(cBuf, "LOGLEVEL") == 0)
        || (strcmp(cBuf, "LOG") == 0)
       )
      uCmd = OP_LOG;

#ifdef AX25IP_DYNLEARN
    /* Timeout fuer dynamische Routen */
    if (strcmp(cBuf, "TIMEOUT") == 0)
      uCmd = OP_TIMEOUT;
#endif

    /* Fehler bei der Auswertung ? */
    if (uCmd == OP_NONE)
    {
      putmsg(inv_mode);
      return;
    }

#ifdef AX25IP_DYNLEARN
    /* Hier rein, wenn der Timeout fuer dynamische Routen geaendert werden soll */
    if (uCmd == OP_TIMEOUT)
    {
      unsigned int uNewTimeout;

      /* Noch was da in der Kommandozeile ? */
      if (!skipsp(&clicnt, &clipoi))
      {
        mbp = getmbp();
        putprintf(mbp, "the actual timeout for dynamic routes is : %u\r", uDynTimeout);
        /* und ab die Post ... */
        prompt(mbp);
        seteom(mbp);
        return;
      }

      /* Timeout lesen */
      uNewTimeout = nxtnum(&clicnt, &clipoi);

      if ((uNewTimeout < 0) || (uNewTimeout > (86400)))  /* max. 1 Tag */
      {
        putmsg("error: timeout out of range (0 - 86400) !\r");
        return;
      }

      /* Wert uebernehmen */
      uDynTimeout = uNewTimeout;

      /* und ab die Post ... */
      mbp = getmbp();
      prompt(mbp);
      seteom(mbp);
      return;
    }
#endif

    /* Hier rein, wenn der Loglevel geaendert werden soll */
    if (uCmd == OP_LOG)
    {
      int new_loglevel;

      /* Noch was da in der Kommandozeile ? */
      if (!skipsp(&clicnt, &clipoi))
      {
        mbp = getmbp();
        putprintf(mbp, "my actual loglevel is : %u\r", loglevel);
        /* und ab die Post ...                                            */
        prompt(mbp);
        seteom(mbp);
        return;
      }

      /* neues Loglevel lesen */
      new_loglevel = nxtnum(&clicnt, &clipoi);

      if ((new_loglevel < 0) || (new_loglevel > 4))
      {
        putmsg("error: loglevel out of range !!!\r");
        return;
      }

      /* Wert uebernehmen */
      loglevel = new_loglevel;

      /* und ab die Post ... */
      mbp = getmbp();
      prompt(mbp);
      seteom(mbp);
      return;
    }

    /* Hier nur rein, wenn wir als naechstes eine UDP-Portnummer erwarten */
    if (uCmd == OP_MYUDP)
    {
      /* Noch was da in der Kommandozeile ? */
      if (!skipsp(&clicnt, &clipoi))
      {
        mbp = getmbp();
        putprintf(mbp, "my actual UDP-port is : %u\r", ntohs(my_udp));
        /* und ab die Post ...                                            */
        prompt(mbp);
        seteom(mbp);
        return;
      }

      /* neue UDP-Portnummer lesen */
      new_udp = nxtnum(&clicnt, &clipoi);

      if (new_udp <= 0 || new_udp >= 65535)
      {
        putmsg("error: UDP-port number not valid !!!\r");
        return;
      }

      /* Laufendes UDP aendern */
#ifndef AXIP_UDPERROR_BUGFIX
      if (fd_udp != -1)
#else
      /* Der Versuch, den Port auf den ohnehin bereits verwendeten Port
         zu aendern, ist kein Fehler sondern wird ignoriert (db1ras)      */
      if ((fd_udp != -1) && (my_udp != htons(new_udp)))
#endif
      {
        /* Versuchen, neuen UDP-Port anzulegen */
        if ((tmp_fd = setup_udp(htons(new_udp))) != -1)
        {
          /* Neuen Descriptor eintragen und alten schliessen */
          close(fd_udp);
          fd_udp = tmp_fd;
          /* Check, ob neuer Filedescriptor groesser ist */
          l1pp->kisslink = max(fd_ip, fd_udp);
          /* Neuen UDP-Port merken */
          my_udp = htons(new_udp);
        }
        else
        {
          putmsg("error: changing the UDP-port failed !!!\r");
          return;
        }
      }
      else
      {
        /* neuen UDP-Port merken */
        my_udp = htons(new_udp);
      }

      /* und ab die Post ...                                                  */
      mbp = getmbp();
      prompt(mbp);
      seteom(mbp);
      return;
    }

    /* Noch was da in der Kommandozeile ? */
    if (!skipsp(&clicnt, &clipoi))
    {
      putmsg("syntax error: callsign missing\r");
      return;
    }

    /* Frischer Buffer */
    memset(cBuf, 0, sizeof(cBuf));

    /* Call lesen */
    for (i = 0; i < BUFLEN; ++i)
    {
      if ((!clicnt) || (*clipoi == ' '))
        break;

      clicnt--;
      cBuf[i] = toupper(*clipoi++);
    }

    /* Operation auf Default-Route ? */
    if (strcmp(cBuf, "DEFAULT") == 0)
      call[0] = NUL;    /* Leere Callstring als Indikator fuer Defaultroute */
    else
    {
      /* Call konvertieren und pruefen */
      if (a_to_call(cBuf, call) != 0)
      {
        putmsg(inv_call);
        return;
      }
    }

    /* Wenn wir loeschen wollen, brauchen wir nichts mehr */
    if (uCmd != OP_DEL)
    {
      /* Noch was da in der Kommandozeile ? */
      if (!skipsp(&clicnt, &clipoi))
      {
        putmsg("syntax error: IP-adress missing\r");
        return;
      }

      /* Frische Buffer */
      memset(hostname, 0, sizeof(hostname));

      /* Hostnamen / IP-Adresse lesen */
      for (i = 0; i < HNLEN; ++i)
      {
        if (!clicnt || *clipoi == ' ')
          break;

        clicnt--;
        hostname[i] = *clipoi++;
      }

      /* eventuellen Hostnamen aufloesen */
      if ((host = gethostbyname(hostname)) == NULL)
      {
        putmsg(inv_ip);
        return;
      }
    } /* if (uCmd != OP_DEL) */

    /* Optionale UDP-Parameter lesen */
    skipsp(&clicnt, &clipoi);

    /* Frischer Buffer */
    memset(cBuf, 0, sizeof(cBuf));

    /* Mode lesen (IP, UDP) */
    for (i = 0; i < BUFLEN; ++i)
    {
      if ((!clicnt) || (*clipoi == ' '))
        break;

      clicnt--;
      cBuf[i] = toupper(*clipoi++);
    }

    /* Mode bestimmen (UDP) */
    /* Bei der Initialisierung wurde bereits IP eingestellt */
    if (   (strcmp(cBuf, "UDP") == 0)
        || (cBuf[0] == 'U')
       )
      uMode = UDP_MODE;

    /* Noch was da in der Kommandozeile ? */
    if (!skipsp(&clicnt, &clipoi))
      port = DEFAULT_UDP_PORT;
    else
    {
      /* neue UDP-Portnummer lesen */
      port = nxtnum(&clicnt, &clipoi);

      if (port <= 0 || port >= 65535)
      {
        putmsg("error: UDP-port number not valid\r");
        return;
      }
    }

    /* Operation ausfuehren */
    switch (uCmd)
    {
      /* Route hinzufuegen */
#ifndef COMPILER_BUGFIX
      case OP_ADD : route_add(host->h_addr_list[0], call, (uMode == UDP_MODE ? port : 0), (call[0] == 0 ? TRUE : FALSE), 0);
#else
      case OP_ADD : route_add((unsigned char *)host->h_addr_list[0], call, (uMode == UDP_MODE ? port : 0), (call[0] == 0 ? TRUE : FALSE), 0);
#endif
                    /*        IP-Adresse          , call (NULL bei Defaultroute), Port (0 bei IP)               , Defaultroute                 , Timeout */
                    break;

      /* Route loeschen */
      case OP_DEL : route_del(call);
                    break;

      default     : break;
    }

    mbp = getmbp();
    /* und ab die Post ...                                                  */
    prompt(mbp);
    seteom(mbp);
    return;
  }

  }
  else
  {
  /* ALTE SYNTAX */

  /* Sysop will aendern und noch was in der Zeile da ? */
  if (issyso() && skipsp(&clicnt, &clipoi))
  {
    clicnt--;

    switch (toupper(*clipoi++))
    {
      case 'P':                                 /* Parser aendern       */
        /* wir koennen nur in den neuen Parser umschalten */
        new_parser = TRUE;
        putmsg("Switched to new axipr-parser !!!\r");
        return;
        break;

      case 'R':                                 /* Routen-Eintrag       */
        /* Kommandozeile pruefen */
        if (!skipsp(&clicnt, &clipoi))
        {
          putmsg(inv_par);
          return;
        }

        clicnt--;

        switch (*clipoi++)
        {
          case '-':                             /* Route loeschen       */
            /* Call lesen */
            if (getcal(&clicnt, &clipoi, TRUE, (char *)call) != YES)
            {
              invcal();
              return;
            }

            /* Call konvertieren */
            for (i = 0; i < L2CALEN; ++i)
              call[i] = call[i] << 1;

            /* Route loeschen */
            route_del(call);
            break;

          case '+':                             /* Route hinzufuegen    */
            /* Call lesen */
            if (getcal(&clicnt, &clipoi, TRUE, (char *)call) != YES)
            {
              invcal();
              return;
            }

            /* Call linksschieben   */
            for (i = 0; i < L2CALEN; ++i)
              call[i] = call[i] << 1;

            /* Noch was da in der Kommandozeile ? */
            if (!skipsp(&clicnt, &clipoi))
            {
              putmsg("IP-Adress/Hostname missing\r");
              return;
            }

            /* Frische Buffer */
            memset(hostname, 0, sizeof(hostname));

            /* Hostnamen / IP-Adresse lesen */
            for (i = 0; i < HNLEN; ++i)
            {
              if (!clicnt || *clipoi == ' ')
                break;

              clicnt--;
              hostname[i] = *clipoi++;
            }

            /* Hostname / IP-Adresse aufloesen */
            if ((host = gethostbyname(hostname)) == NULL)
            {
              putmsg(inv_ip);
              return;
            }

            /* Mode-Parameter vorhanden ? */
            if (skipsp(&clicnt, &clipoi))
            {
              --clicnt;
              switch (toupper(*clipoi++))
              {
                case 'I':       /* IP: Port steht fest */
                  break;
                case 'U':       /* UDP: abweichenden Port lesen wenn vorhanden */
                  nextspace(&clicnt, &clipoi);
                  if (!skipsp(&clicnt, &clipoi))
                    port = DEFAULT_UDP_PORT;
                  else
                    port = nxtnum(&clicnt, &clipoi);
                  break;

                default:        /* unbekannter Modus */
                  putmsg(inv_mode);
                  return;
              }
            }
            /* Route hinzufuegen */
            route_add((unsigned char *)host->h_addr_list[0], call, (int) port, FALSE, 0);
            break;
        }
        break;

      case 'D':                         /* Default Route aendern        */

        nextspace(&clicnt, &clipoi);

        /* Noch Eingaben da ? */
        if (!skipsp(&clicnt, &clipoi))
        {
          putmsg(inv_mode);
          return;
        }

        clicnt--;

        /* Operation bestimmen */
        switch (*clipoi++)
        {
          case '-':                     /* Default Route loeschen       */
            route_del((unsigned char *)"");
            break;
          case '+':                     /* Default Route eingeben       */

            /* IP-Adresse / Hostname lesen */
            if (!skipsp(&clicnt, &clipoi))
            {
              putmsg(inv_ip);
              return;
            }

            /* Frische Buffer */
            memset(hostname, 0, sizeof(hostname));

            /* IP-Adresse / Hostname kopieren */
            for (i = 0; i < HNLEN; ++i)
            {
              if (!clicnt || *clipoi == ' ')
                break;
              clicnt--;
              hostname[i] = *clipoi++;
            }

            /* aufloesen */
            if ((host = gethostbyname(hostname)) == NULL)
            {
              putmsg(inv_ip);
              return;
            }

            /* Modus-Kennzeichner lesen falls vorhanden */
            if (skipsp(&clicnt, &clipoi))
            {
              --clicnt;
              switch (toupper(*clipoi++))
              {
                case 'I':   /* IP: Port steht fest */
                  break;
                case 'U':   /* UDP: eventuell abweichenden Port lesen */
                  nextspace(&clicnt, &clipoi);

                  if (!skipsp(&clicnt, &clipoi))
                    port = DEFAULT_UDP_PORT;
                  else
                    port = nxtnum(&clicnt, &clipoi);
                  break;
                default:    /* unbekannter Modus */
                  putmsg(inv_mode);
                  return;
              }
            }
            /* Route eintragen */
            route_add((unsigned char *)host->h_addr_list[0], NULL, (int) port, TRUE, 0);
            break;
          default:
            putmsg(inv_mode);
            return;
        }
        break;

      /* Eigenen UDP-Port aendern */
      case 'U':
        /* neuer Port muss vorhanden sein */
        if (!skipsp(&clicnt, &clipoi))
        {
          putmsg(inv_par);
          return;
        }

        /* Port pruefen */
        if ((new_udp = nxtnum(&clicnt, &clipoi)) <= 0)
        {
          putmsg("UDP-Port not valid, not changed !!!\r");
          return;
        }

        /* Laufendes UDP aendern */
        if (fd_udp != -1)
        {
          /* Versuchen, neuen UDP-Port anzulegen */
          if ((tmp_fd = setup_udp(htons(new_udp))) != -1)
          {
            /* Neuen Descriptor eintragen und alten schliessen */
            close(fd_udp);
            fd_udp = tmp_fd;
            /* Check, ob neuer Filedescriptor groesser ist */
            l1pp->kisslink = max(fd_ip, fd_udp);
            /* Neuen UDP-Port merken */
            my_udp = htons(new_udp);
          }
          else
          {
            putmsg("ERROR: Changing UDP-Port failed, Port not changed !!!\r");
            return;
          }
        }
        else
        {
          /* neuen UDP-Port merken */
          my_udp = htons(new_udp);
        }

        putmsg("UDP-Port successfully changed\r");
        break;

      /* Unbekannte Kommandos */
      default:
        putmsg(inv_par);
        return;
    }
  }
  }

/* TEST DG9OBU */
#ifndef AX25IP_DYNLEARN
  mbp = putals("AXIP-Routes:\rCall-------IP----------------Mode--Port-\r");
#else
  mbp = putals("AXIP-Routes:\rCall-------IP----------------Mode--Port--Timeout\r");
#endif

  /* Alle Routingtabelleneintraege durchgehen und ausgeben */
  for (i = 0; i < route_tbl_top; ++i)
    show_rt_entry(route_tbl[i], mbp);

  /* Default-Route ausgeben wenn vorhanden */
  if (default_route.callsign[0] == 1)
    show_rt_entry(default_route, mbp);

  putprintf(mbp, "------------------------------------------------\r");

  /* Eigenen UDP-Port ausgeben wenn UDP verwendet */
  if (fd_udp != -1)
    putprintf(mbp, "I am listening on UDP-port %u\r", ntohs(my_udp));

#ifdef AX25IP_DYNLEARN
  /* Timeout des dynamischen Routenlerners ausgeben */
  putprintf(mbp, "Timeout for dynamically learned routes is %u seconds\r", uDynTimeout);
#endif

  /* und ab die Post ... */
  prompt(mbp);
  seteom(mbp);
}

/* Routeneintraege fuer parms.tnb ausgeben */
void dump_ax25ip(MBHEAD* mbp)
{
  register int i;

  putstr(";\r; AX25IP Routes\r;\r", mbp);

  if (new_parser == TRUE)
  {
    putstr("; WARNING: this section is written using the new syntax of the AXIPR-command,\r", mbp);
    putstr(";          DO NOT USE with older TNNs compiled only for old style syntax !!!\r;\r", mbp);

    /* Falls wir noch im alten Parser sind, dann umschalten */
    putprintf(mbp, "AXIPR P\r");
    /* eigenen UDP-Port ausgeben */
    putprintf(mbp, "AXIPR MYUDP %u\r", ntohs(my_udp));
#ifdef AX25IP_DYNLEARN
    /* Timeout fuer dynamische Routen ausgeben */
    putprintf(mbp, "AXIPR TIMEOUT %u\r", uDynTimeout);
#endif
  }

  /* alle Routingtabelleneintraege durchgehen und ausgeben */
  for (i = 0; i < route_tbl_top; ++i)
  {
     /* TEST DG9OBU */
#ifdef AX25IP_DYNLEARN
    /* Dynamisch gelernte Routen werden nicht geschrieben ! */
    if (route_tbl[i].timeout != 0)
      continue;
#endif

    /* Eintrag einer Route erzeugen */
    putprintf(mbp, (new_parser == FALSE ? "AXIPR R + %s " : "AXIPR ADD %s ")
                 , call_to_a(&route_tbl[i].callsign[0]));

    /* Die IP-Adresse ausgeben */
    putprintf(mbp, "%d.", route_tbl[i].ip_addr[0]);
    putprintf(mbp, "%d.", route_tbl[i].ip_addr[1]);
    putprintf(mbp, "%d.", route_tbl[i].ip_addr[2]);
    putprintf(mbp, "%d",  route_tbl[i].ip_addr[3]);

    /* Mode und ggf. Port ausgeben wenn UDP */
    if (route_tbl[i].udp_port != 0)
    {
      putstr((new_parser == FALSE ? " U " : " UDP "), mbp);
      putnum(ntohs(route_tbl[i].udp_port), mbp);
    }

    putstr("\r", mbp);
  }

  /* Die Defaultroute ausgeben wenn gesetzt */
  if (default_route.callsign[0] == 1)
  {
    /* Kommandokopf schreiben */
    putstr((new_parser == FALSE ? "AXIPR D + " : "AXIPR ADD DEFAULT "), mbp);

    /* IP-Adresse ausgeben */
    putprintf(mbp, "%d.", default_route.ip_addr[0]);
    putprintf(mbp, "%d.", default_route.ip_addr[1]);
    putprintf(mbp, "%d.", default_route.ip_addr[2]);
    putprintf(mbp, "%d",  default_route.ip_addr[3]);

    /* Mode und ggf. Port ausgeben wenn UDP */
    if (default_route.udp_port != 0)
    {
      putstr((new_parser == FALSE ? " U " : " UDP "), mbp);
      putnum(ntohs(default_route.udp_port), mbp);
    }
    putstr("\r", mbp);
  }
}

/************************************************************************/
/* AX25IP Initialisierung                                               */
/************************************************************************/
BOOLEAN ax25ip_l1init(int l2port)
{
  if (ax25ip_active == TRUE)
    return(TRUE);

  /* Socketstrukturen initialisieren */
  memset((char*)&to, 0, sizeof(to));
  memset((char*)&from, 0, sizeof(from));

  ax25ip_port = l2port;
  l1pp = &l1port[l1ptab[l2port]];
  l2flp = (LHEAD *) &txl2fl[l2port];

  /* Routingtabelle initialisieren */
  route_init();

  /* Konfiguration einlesen */
#ifndef AXIP_CFG_BUGFIX
  if (config_read() == FALSE)
    return(FALSE);
#else
  /* Fehler in der Konfiguration werden ignoriert (db1ras) */
  config_read();
#endif

  /* Notwendige Sockets anlegen */
  ax25ip_check_up();

  /* Wir sind fertig */
  ax25ip_active = TRUE;
  return(TRUE);
}

/* Einen Filedescriptor fuer IP anlegen */
int setup_ip(void)
{
  /* Temporaerer Filedescriptor */
  int tmp_fd = -1;

  /* Adressstruktur loeschen */
  memset((char*)&ipbind, 0, sizeof(ipbind));

  /* Rohen IP-Socket anlegen */
  if ((tmp_fd = socket(AF_INET, SOCK_RAW, IPPROTO_AX25)) < 0)
  {
    xprintf("AX25IP : cannot create ip raw-socket: %s\n", strerror(errno));
    LOGL2("AX25IP : cannot create ip raw-socket: %s\n", strerror(errno));
    return(-1);
  }

  /* Nonblocking-IO */
  if (fcntl(tmp_fd, F_SETFL, FNDELAY) < 0)
  {
    xprintf("AX25IP : cannot set non-blocking I/O on ip raw-socket\n");
    LOGL2("AX25IP : cannot set non-blocking I/O on ip raw-socket\n");
    close(tmp_fd);
    return(-1);
  }

  /* Socket-Optionen fuer RAW-IP einstellen */
  /* SO_KEEPALIVE einschalten */
  if (sockopt_keepalive == TRUE)
  {
    int iFlag = 1;

    if (setsockopt(tmp_fd, SOL_SOCKET, SO_KEEPALIVE, &iFlag, sizeof(int)) < 0)
    {
      xprintf("AX25IP : cannot set SO_KEEPALIVE for ip : %s\n", strerror(errno));
      LOGL2("AX25IP : cannot set SO_KEEPALIVE for ip : %s\n", strerror(errno));
    }
  }

  /* Socket auf max. Durchsatz */
  if (sockopt_throughput == TRUE)
  {
    int iFlag = IPTOS_THROUGHPUT;

    if (setsockopt(tmp_fd, IPPROTO_IP, IP_TOS, &iFlag, sizeof(IPTOS_THROUGHPUT)) < 0)
    {
      xprintf("AX25IP : cannot set IPTOS_THROUGHPUT for ip : %s\n", strerror(errno));
      LOGL2("AX25IP : cannot set IPTOS_THROUGHPUT for ip : %s\n", strerror(errno));
    }
  }

  /* Adressstruktur mit notwendigen Werten fuellen */
  ipbind.sin_family = AF_INET;
  ipbind.sin_addr.s_addr = INADDR_ANY;

  return (tmp_fd);
}

/* Einen Filedescriptor fuer UDP anlegen */
/* Es muss der UDP-Port angegeben werden, auf dem wir hoeren wollen */
/* (Achtung: UDP-Port in Network-Byteorder) */
int setup_udp(unsigned short udp_port)
{
  /* Temporaerer Filedescriptor */
  int tmp_fd = -1;

  /* Port-Check */
  if (udp_port == 0)
    return(-1);

  /* Adressstruktur loeschen */
  memset((char*)&udpbind, 0, sizeof(udpbind));

  /* UDP-Socket anlegen */
  if ((tmp_fd = socket(AF_INET, SOCK_DGRAM, 0)) < 0)
  {
    xprintf("AX25IP : cannot create socket: %s\n", strerror(errno));
    LOGL2("AX25IP : cannot create socket: %s\n", strerror(errno));
    return(-1);
  }

  /* Nonblocking-IO */
  if (fcntl(tmp_fd, F_SETFL, FNDELAY) < 0)
  {
    xprintf("AX25IP : cannot set non-blocking I/O on udp socket\n");
    LOGL2("AX25IP : cannot set non-blocking I/O on udp socket\n");
    close(tmp_fd);
    return(-1);
  }

  /* Socket-Optionen fuer UDP einstellen */
  /* SO_KEEPALIVE einschalten */
  if (sockopt_keepalive == TRUE)
  {
    int iFlag = 1;

    if (setsockopt(tmp_fd, SOL_SOCKET, SO_KEEPALIVE, &iFlag, sizeof(int)) < 0)
    {
      xprintf("AX25IP : cannot set SO_KEEPALIVE for udp : %s\n", strerror(errno));
      LOGL2("AX25IP : cannot set SO_KEEPALIVE for udp : %s\n", strerror(errno));
    }
  }

  /* Socket auf max. Durchsatz */
  if (sockopt_throughput == TRUE)
  {
    int iFlag = IPTOS_THROUGHPUT;

    if (setsockopt(tmp_fd, IPPROTO_IP, IP_TOS, &iFlag, sizeof(IPTOS_THROUGHPUT)) < 0)
    {
      xprintf("AX25IP : cannot set IPTOS_THROUGHPUT for udp : %s\n", strerror(errno));
      LOGL2("AX25IP : cannot set IPTOS_THROUGHPUT for udp : %s\n", strerror(errno));
    }
  }

  /* Adressstruktur fuellen */
  udpbind.sin_addr.s_addr = INADDR_ANY;
  udpbind.sin_port = udp_port;
  udpbind.sin_family = AF_INET;

  /* Adresse und Port binden */
  if (bind(tmp_fd, (struct sockaddr *)&udpbind, sizeof(udpbind)) < 0)
  {
    xprintf("AX25IP : cannot bind udp socket: %s\n", strerror(errno));
    LOGL2("AX25IP : cannot bind udp socket: %s\n", strerror(errno));
    close(tmp_fd);
    return(-1);
  }

  return(tmp_fd);
}

/* Pruefen, ob schon Sockets angelegt sind, wenn nicht, dann notwendige */
/* Sockets oeffnen */
void ax25ip_check_up(void)
{
  /* IP notwendig ? */
  if ((count_routes(FALSE) != 0) && (fd_ip == -1))
    fd_ip = setup_ip(); /* IP starten */

  /* UDP notwendig ? */
  if ((count_routes(TRUE) != 0) && (fd_udp == -1))
    fd_udp = setup_udp(my_udp); /* UDP starten */

  /* Markieren, ob IP oder UDP laufen */
  l1pp->kisslink = max(fd_ip, fd_udp);
}

/* Pruefen, ob Sockets angelegt sind und diese noch benoetigt werden. */
/* Falls nicht, Sockets schliessen */
void ax25ip_check_down(void)
{
  /* IP notwendig ? */
  if ((count_routes(FALSE) == 0) && (fd_ip != -1))
  {
    close(fd_ip);
    fd_ip = -1;
  }

  /* UDP notwendig ? */
  if ((count_routes(TRUE) == 0) && (fd_udp != -1))
  {
    close(fd_udp);
    fd_udp = -1;
  }

  /* Markieren, ob IP oder UDP laufen */
  l1pp->kisslink = max(fd_ip, fd_udp);
}

/*************************************************************************/
/* AX25IP Pakete empfangen                                               */
/*************************************************************************/
void ax25ip_recv(void)
{
  struct       ip *ipptr;
  struct  timeval tv;
  int          l = 0;
  int          iDescriptorsReady = 0;
  register int i = 0;
  register int max_fd = -1;
  socklen_t    fromlen = sizeof(from);
  int          hdr_len = 0;
  UBYTE        buf[MAX_FRAME+1];
  UBYTE        *bufptr;
  MBHEAD       *rxfhd;
  BOOLEAN      UDP_Frame = FALSE;
  fd_set       rmask;

  FD_ZERO(&rmask);

  /* IP-Filedescriptor eintragen */
  if (fd_ip != -1)
  {
    FD_SET(fd_ip, &rmask);
    if (fd_ip > max_fd - 1)
      max_fd = fd_ip + 1;
  }

  /* UDP-Filedescriptor eintragen */
  if (fd_udp != -1)
  {
    FD_SET(fd_udp, &rmask);
    if (fd_udp > max_fd - 1)
      max_fd = fd_udp + 1;
  }

  /* ist was aktiv ? */
  if (max_fd == -1)
    return;

  tv.tv_usec = 0;
  tv.tv_sec = 0;

  iDescriptorsReady = select(max_fd, &rmask, NULL, NULL, &tv);

  /* nix da */
  if (iDescriptorsReady == 0)
    return;

  /* Fehler */
  if (iDescriptorsReady == -1)
  {
    LOGL2("select()-Error %i: %s", errno, strerror(errno));
    return;
  }

  /* RX fuer IP */
  if ((fd_ip != -1) && (FD_ISSET(fd_ip, &rmask)))
  {
    l = recvfrom(fd_ip, (UBYTE *)(bufptr = buf), MAX_FRAME, 0, (struct sockaddr *) &from, &fromlen);

    if (l > 0)
    {
      if (l > sizeof(struct ip))
      {
        ipptr = (struct ip*)buf;
        hdr_len = 4 * ipptr->ip_hl;

        if (!ok_crc(buf + hdr_len, l - hdr_len))
        { /* stimmt die CRC ? */
          LOGL2("IP-RX: CRC-Error, frame dropped");
          return; /* Fehler */
        }
      }
      else
      {
        LOGL2("IP-RX: frame too short, frame dropped");
        return; /* Fehler */
      }
    }
    i = hdr_len;
  }

  /* RX fuer UDP, aber nur, wenn nicht schon TCP was empfangen hat */
  if ((fd_udp != -1) && (FD_ISSET(fd_udp, &rmask) && (l == 0)))
  {
    l = recvfrom(fd_udp, (UBYTE *)(bufptr = buf), MAX_FRAME, 0, (struct sockaddr *) &from, &fromlen);

    if (l > 0)
    {
      if (!ok_crc(buf, l))
      { /* stimmt die CRC ? */
        LOGL2("UDP-RX: CRC-Error, frame dropped");
        return; /* Fehler */
      }
    }
    UDP_Frame = TRUE;
  }

  /* Ab hier gemeinsame Behandlung */
  /* Erst mal gucken, ob was gekommen ist */
  if (l == 0)
    return; /* nix da */

  l -= 2; /* CRC abschneiden */

  if (l < 15)
  {
    LOGL2("AX25IP: RX frame dropped, length wrong, frame is too short !");
    return;
  }

  /* Hier angekommen haben wir (hoffentlich) ein gueltiges Frame. Die */
  /* weitere Behandlung ist fuer IP und UDP nahezu identisch.         */

  /* Puffer holen und Empfangsport eintragen */
  (rxfhd = (MBHEAD *)allocb(ALLOC_MBHEAD))->l2port = ax25ip_port;

  /* Frame in Buffer umkopieren */
  while (i < l)
    putchr(buf[i++], rxfhd);

  /* Frame zum L2 geben */
  relink((LEHEAD *)rxfhd, (LEHEAD *)rxfl.tail);

  /* TEST DG9OBU */
#ifdef AX25IP_DYNLEARN
  /* Frame an Routenlerner zur Analyse */
  if (route_canlearn(rxfhd))
  {
    /* Frame kann zum Lernen benutzt werden */
    /* Absenderrufzeichen ermitten und konvertieren       */
    /* Wir brauchen hier nicht den eigentlichen Absender, */
    /* sondern den, der als letztes gedigipeated hat.     */
    unsigned char *ipptr;
    unsigned char srccall[L2IDLEN + 1];
    cpyid((char *)srccall, dheardcall(rxfhdr));

    /* Call linksschieben */
    for (i = 0; i < L2CALEN; ++i)
      srccall[i] = srccall[i] << 1;

    /* Gibt es schon eine IP fuer dieses Call, dann kann sich die IP */
    /* geaendert haben, alten Eintrag loeschen und neuen machen. Die */
    /* Default-Route wird allerdings nie geaendert !!!               */
    if (((ipptr = call_to_ip(srccall)) != NULL) && (ipptr != default_route.ip_addr))
      route_del(srccall);

    /* Route eintragen */
    route_add((unsigned char *)&from.sin_addr.s_addr, srccall,
              (UDP_Frame == TRUE ? ntohs(from.sin_port) : 0), FALSE, uDynTimeout);
  }
#endif
}

/*************************************************************************/
/* AX25IP Pakete senden                                                  */
/*************************************************************************/
void ax25ip_send(void)
{
  unsigned char  buf[MAX_FRAME];
  int            len = 0;
  int            fd = -1;
  MBHEAD        *txfhdl;
  unsigned char *targetip;
  unsigned char *call;

  if (kick[ax25ip_port]) /* haben wir was zu senden ? */
  {
    /* Sicherheitsabfrage, da kick[] auch manipuliert werden kann */
    if (l2flp->head == l2flp)
    {
      kick[ax25ip_port] = FALSE;
      return;
    }

    ulink((LEHEAD *)(txfhdl = (MBHEAD *) l2flp->head));/*Zeiger holen*/

    while (    (txfhdl->mbgc < txfhdl->mbpc) /* solange Daten vorhanden sind */
            && (len < MAX_FRAME))            /* aber nur so viel wie Buffer  */
        buf[len++] = getchr(txfhdl);

    relink((LEHEAD *)txfhdl,          /* als gesendet betrachten und in */
           (LEHEAD *)stfl.tail);      /* die gesendet Liste umhaengen   */

    kick[ax25ip_port] = ((LHEAD *)l2flp->head != l2flp);

    /* Naechstes Call der Digipeaterkette holen */
    call = next_addr(buf);

    /* IP-Adresse des naechsten Digipeaters ermitteln */
    targetip = call_to_ip(call);

    if (targetip == NULL) /* wir kennen Ziel nicht */
    {
      LOGL2("unknown Target-IP-Adress %s", call_to_a(call));
      return;
    }

    /* Checksumme anhaengen */
    add_ax25crc(buf, len);
    /* CRC-Bytes sind dazugekommen */
    len += 2;
    /* Ziel-IP-Adresse in Sendestruktur eintragen */
    memcpy((char *) &to.sin_addr, targetip, 4);
    /* Ziel-Port (UDP) eintragen */
    memcpy((char *) &to.sin_port, &targetip[4], 2);
    to.sin_family = AF_INET;

    /* Unterscheidung IP oder UDP */
    if (to.sin_port == 0)
      fd = fd_ip;
    else
      fd = fd_udp;

    if (fd == -1)
    {
      LOGL2("error: no descriptor, can't send !!!\n");
      return;
    }

    if (sendto(fd, (char *)buf, len, 0, (struct sockaddr *)&to, sizeof(to)) < 0)
      LOGL2("error: sendto failed, fd=%u error=%s", fd, strerror(errno));
  }
}

/*************************************************************************/
/* AX25IP Deinitialisierung, Port und Sockets schliessen                 */
/*************************************************************************/
void ax25ip_l1exit(void)
{
  /* Nur wenn aktiv */
  if (ax25ip_active == FALSE)
    return;

  /* IP-Socket schliessen wenn offen */
  if (fd_ip != -1)
  {
    close(fd_ip);
    fd_ip = -1;
  }

  /* UDP-Socket schliessen wenn offen */
  if (fd_udp != -1)
  {
    close(fd_udp);
    fd_udp = -1;
  }

  /* AX25IP nicht mehr aktiv */
  ax25ip_active = FALSE;
  l1pp->kisslink = -1;
}

void ax25ip_l1ctl(int req, int port)
{
  /* nur eigene Ports bearbeiten */
  if ((ax25ip_active == FALSE) || (kissmode(port) != KISS_AXIP))
    return;

  switch (req)
  {
    /* Testpattern auf dem Port senden (wird unterdrueckt) */
    case L1CTST : testflag[port] = FALSE;
                  kick[port] = FALSE;
                  break;

    default : break;
  }
}

void ax25ip_hwstr(int port, MBHEAD *mbp)
{
}

BOOLEAN ax25ip_dcd(int port)
{
  return (FALSE);
}


/* routing.c    Routing table manipulation routines
 *
 * Copyright 1991, Michael Westerhof, Sun Microsystems, Inc.
 * This software may be freely used, distributed, or modified, providing
 * this header is not removed.
 *
 */



/* Initialize the routing table */
void route_init(void)
{
  default_route.callsign[0] = '\0';    /* set to non-zero if valid */
}

/* Add a new route entry */
static void route_add(unsigned char *ip, unsigned char *call, int udpport,
                      int default_rt, int timeout)
{
   register unsigned int i = 0;
   unsigned char *ipptr;

   /* Keine Eintragung ohne IP-Adresse */
   if (ip == NULL)
     return;

   /* Eintrag soll Defaultroute sein */
   if (default_rt)
   {
      for (i = 1; i < 7 ;++i)
        default_route.callsign[i] = 0;

      default_route.padcall = 0;
      (void)memcpy(default_route.ip_addr, ip, 4);
      default_route.udp_port = htons(udpport);
      default_route.pad1 = 0;
      default_route.pad2 = 0;
/* TEST DG9OBU */
#ifdef AX25IP_DYNLEARN
      default_route.timeout = 0;                /* Defaultroute ohne Timeout */
#endif
      default_route.callsign[0] = 1;            /* mark valid */

#ifdef AXIP_UP_BUGFIX
      /* Socket nur installieren, wenn der Port bereits initialisiert wurde
         (db1ras)                                                            */
      if (ax25ip_active == TRUE)
#endif
      /* Socket ggf. installieren */
      ax25ip_check_up();

      return;
   }

   /* Normale Routen nur wenn Call vorhanden */
   if (call == NULL)
     return;

   /* Check auf doppelte Eintraege, Eintrag wird nur angenommen wenn */
   /* es noch noch keinen Eintrag unter diesem Call gibt. */
   if (((ipptr = call_to_ip(call)) != NULL) && (ipptr != default_route.ip_addr))
   {
#ifndef AXIP_PUTMSG_BUGFIX
     putmsg("A route to this callsign is already set up, delete it first.\n");
#else
     /* \r ist das korrekte Zeilenende fuer putmsg (db1ras) */
     putmsg("A route to this callsign is already set up, delete it first.\r");
#endif
     return;
   }

   /* Passt noch ein Eintrag in die Tabelle ? */
   if (route_tbl_top >= TABLE_SIZE)
   {
     xprintf("Routing table is full; entry ignored.\n");
     LOGL2("Routing table is full; entry ignored");
#ifndef AXIP_PUTMSG_BUGFIX
     putmsg("Routing table is full, entry ignored\n");
#else
     /* \r ist das korrekte Zeilenende fuer putmsg (db1ras) */
     putmsg("Routing table is full, entry ignored\r");
#endif
     return;
   }

   /* Call eintragen */
   for (i = 0; i < 6; ++i)
     route_tbl[route_tbl_top].callsign[i] = call[i] & 0xfe;

   /* SSID eintragen */
   route_tbl[route_tbl_top].callsign[6] = (call[6] & 0x1e) | 0x60;

   route_tbl[route_tbl_top].padcall = 0;

   /* IP-Adresse und bei UDP den UDP-Port eintragen */
   (void)memcpy(route_tbl[route_tbl_top].ip_addr, ip, 4);
   route_tbl[route_tbl_top].udp_port = htons(udpport);

   /* Padding */
   route_tbl[route_tbl_top].pad1 = 0;
   route_tbl[route_tbl_top].pad2 = 0;

/* TEST DG9OBU */
#ifdef AX25IP_DYNLEARN
   route_tbl[route_tbl_top].timeout = timeout;
#endif

   /* Ein Routeneintrag mehr */
   route_tbl_top++;

#ifdef AXIP_UP_BUGFIX
    /* Socket nur installieren, wenn der Port bereits initialisiert wurde
       (db1ras)                                                              */
    if (ax25ip_active == TRUE)
#endif
   /* Socket ggf. installieren */
   ax25ip_check_up();

   return;
}

/*
 * Return an IP address and port number given a callsign.
 * We return a pointer to the address; the port number can be found
 * immediately following the IP address. (UGLY coding; to be fixed later!)
 */

unsigned char *call_to_ip(unsigned char *call)
{
   register int i;
   unsigned char mycall[7];

   /* Leere Calls koennen nicht verarbeitet werden */
   if (call == NULL)
     return(NULL);

   /* Call-Bytes kopieren, letztes Bit abschneiden (entspricht >> 1 & 7F) */
   for (i = 0; i < 6; ++i)
     mycall[i] = call[i] & 0xfe;

   /* SSID kopieren */
   mycall[6] = (call[6] & 0x1e) | 0x60;

   /* Routingtabelle durchsuchen */
   for (i = 0; i < route_tbl_top; ++i)
     if (addrmatch(mycall , route_tbl[i].callsign))
     {
#ifdef AX25IP_DYNLEARN
       /* Aktivitaet auf diesem Eintrag, dann Timeout erneuern */
       if (route_tbl[i].timeout)
         route_tbl[i].timeout = uDynTimeout;
#endif
       return(route_tbl[i].ip_addr); /* Eintrag gefunden */
     }

   /* Kein Routingeintrag vorhanden, Default-Route melden wenn vorhanden */
   if (default_route.callsign[0])
      return(default_route.ip_addr);

   /* Kein Routingeintrag gefunden, keine Defaultroute */
   return(NULL);
}

/*
 * tack on the CRC for the frame.  Note we assume the buffer is long
 * enough to have the two bytes tacked on.
 */
void add_ax25crc(unsigned char *buf, int l)
{
   unsigned short int u = 0;

   u = compute_crc(buf, l);

   buf[l] = u & 0xff;              /* lsb first */
   buf[l + 1] = (u >> 8) & 0xff;   /* msb next */
}


/*
 **********************************************************************
 * The following code was taken from Appendix B of RFC 1171
 * (Point-to-Point Protocol)
 *
 * The RFC credits the following sources for this implementation:
 *
 *   Perez, "Byte-wise CRC Calculations", IEEE Micro, June, 1983.
 *
 *   Morse, G., "Calculating CRC's by Bits and Bytes", Byte,
 *   September 1986.
 *
 *   LeVan, J., "A Fast CRC", Byte, November 1987.
 *
 *
 * The HDLC polynomial: x**0 + x**5 + x**12 + x**16
 */

/*
 * u16 represents an unsigned 16-bit number.  Adjust the typedef for
 * your hardware.
 */
typedef unsigned short u16;


/*
 * FCS lookup table as calculated by the table generator in section 2.
 */
static u16 fcstab[256] = {
   0x0000, 0x1189, 0x2312, 0x329b, 0x4624, 0x57ad, 0x6536, 0x74bf,
   0x8c48, 0x9dc1, 0xaf5a, 0xbed3, 0xca6c, 0xdbe5, 0xe97e, 0xf8f7,
   0x1081, 0x0108, 0x3393, 0x221a, 0x56a5, 0x472c, 0x75b7, 0x643e,
   0x9cc9, 0x8d40, 0xbfdb, 0xae52, 0xdaed, 0xcb64, 0xf9ff, 0xe876,
   0x2102, 0x308b, 0x0210, 0x1399, 0x6726, 0x76af, 0x4434, 0x55bd,
   0xad4a, 0xbcc3, 0x8e58, 0x9fd1, 0xeb6e, 0xfae7, 0xc87c, 0xd9f5,
   0x3183, 0x200a, 0x1291, 0x0318, 0x77a7, 0x662e, 0x54b5, 0x453c,
   0xbdcb, 0xac42, 0x9ed9, 0x8f50, 0xfbef, 0xea66, 0xd8fd, 0xc974,
   0x4204, 0x538d, 0x6116, 0x709f, 0x0420, 0x15a9, 0x2732, 0x36bb,
   0xce4c, 0xdfc5, 0xed5e, 0xfcd7, 0x8868, 0x99e1, 0xab7a, 0xbaf3,
   0x5285, 0x430c, 0x7197, 0x601e, 0x14a1, 0x0528, 0x37b3, 0x263a,
   0xdecd, 0xcf44, 0xfddf, 0xec56, 0x98e9, 0x8960, 0xbbfb, 0xaa72,
   0x6306, 0x728f, 0x4014, 0x519d, 0x2522, 0x34ab, 0x0630, 0x17b9,
   0xef4e, 0xfec7, 0xcc5c, 0xddd5, 0xa96a, 0xb8e3, 0x8a78, 0x9bf1,
   0x7387, 0x620e, 0x5095, 0x411c, 0x35a3, 0x242a, 0x16b1, 0x0738,
   0xffcf, 0xee46, 0xdcdd, 0xcd54, 0xb9eb, 0xa862, 0x9af9, 0x8b70,
   0x8408, 0x9581, 0xa71a, 0xb693, 0xc22c, 0xd3a5, 0xe13e, 0xf0b7,
   0x0840, 0x19c9, 0x2b52, 0x3adb, 0x4e64, 0x5fed, 0x6d76, 0x7cff,
   0x9489, 0x8500, 0xb79b, 0xa612, 0xd2ad, 0xc324, 0xf1bf, 0xe036,
   0x18c1, 0x0948, 0x3bd3, 0x2a5a, 0x5ee5, 0x4f6c, 0x7df7, 0x6c7e,
   0xa50a, 0xb483, 0x8618, 0x9791, 0xe32e, 0xf2a7, 0xc03c, 0xd1b5,
   0x2942, 0x38cb, 0x0a50, 0x1bd9, 0x6f66, 0x7eef, 0x4c74, 0x5dfd,
   0xb58b, 0xa402, 0x9699, 0x8710, 0xf3af, 0xe226, 0xd0bd, 0xc134,
   0x39c3, 0x284a, 0x1ad1, 0x0b58, 0x7fe7, 0x6e6e, 0x5cf5, 0x4d7c,
   0xc60c, 0xd785, 0xe51e, 0xf497, 0x8028, 0x91a1, 0xa33a, 0xb2b3,
   0x4a44, 0x5bcd, 0x6956, 0x78df, 0x0c60, 0x1de9, 0x2f72, 0x3efb,
   0xd68d, 0xc704, 0xf59f, 0xe416, 0x90a9, 0x8120, 0xb3bb, 0xa232,
   0x5ac5, 0x4b4c, 0x79d7, 0x685e, 0x1ce1, 0x0d68, 0x3ff3, 0x2e7a,
   0xe70e, 0xf687, 0xc41c, 0xd595, 0xa12a, 0xb0a3, 0x8238, 0x93b1,
   0x6b46, 0x7acf, 0x4854, 0x59dd, 0x2d62, 0x3ceb, 0x0e70, 0x1ff9,
   0xf78f, 0xe606, 0xd49d, 0xc514, 0xb1ab, 0xa022, 0x92b9, 0x8330,
   0x7bc7, 0x6a4e, 0x58d5, 0x495c, 0x3de3, 0x2c6a, 0x1ef1, 0x0f78
};

#define PPPINITFCS      0xffff  /* Initial FCS value */
#define PPPGOODFCS      0xf0b8  /* Good final FCS value */

/*
 * Calculate a new fcs given the current fcs and the new data.
 */
u16 pppfcs(register u16 fcs, register unsigned char *cp, register int len)
{
    while (len--)
       fcs = (fcs >> 8) ^ fcstab[(fcs ^ *cp++) & 0xff];

    return(fcs);
}

/*
 * End code from Appendix B of RFC 1171
 **********************************************************************
 */

/*
 *  The following routines are simply convenience routines...
 *  I'll merge them into the mainline code when suitably debugged
 */

/* Return the computed CRC */
unsigned short int compute_crc(unsigned char *buf, int l)
{
   int fcs = 0;

   fcs = PPPINITFCS;
   fcs = pppfcs(fcs, buf, l);
   fcs ^= 0xffff;
   return(fcs);
}

/* Return true if the CRC is correct */
int ok_crc(unsigned char *buf, int l)
{
   int fcs = 0;

   fcs = PPPINITFCS;
   fcs = pppfcs(fcs, buf, l);
   return(fcs == PPPGOODFCS);
}

/* return true if the addresses supplied match */
int addrmatch(unsigned char *a, unsigned char *b)
{
   if ((*a == '\0') || (*b == '\0')) return(0);

   if ((*a++ ^ *b++) & 0xfe) return(0);   /* "K" */
   if ((*a++ ^ *b++) & 0xfe) return(0);   /* "A" */
   if ((*a++ ^ *b++) & 0xfe) return(0);   /* "9" */
   if ((*a++ ^ *b++) & 0xfe) return(0);   /* "W" */
   if ((*a++ ^ *b++) & 0xfe) return(0);   /* "S" */
   if ((*a++ ^ *b++) & 0xfe) return(0);   /* "B" */
   if ((*a++ ^ *b++) & 0x1e) return(0);   /* ssid */
   return(1);
}

/* return pointer to the next station to get this packet */
unsigned char *next_addr(unsigned char *f)
{
   unsigned char *a;

   /* If no digis, return the destination address */
   if (NO_DIGIS(f)) return(f);

   /* check each digi field.  The first one that hasn't seen it is the one */
   a = f + 7;
   do {
      a += 7;
      if (NOTREPEATED(a)) return(a);
   } while (NOT_LAST(a));

/* all the digis have seen it.  return the destination address */
   return(f);
}

/* Open and read the config file */
int config_read(void)
{
   FILE *cf;
   char buf[256], cbuf[256];
#ifndef AXIP_CFG_BUGFIX
   int errflag = 0;
#endif
   int e = 0;
   int lineno = 0;
   char cfgfile[256];

   strcpy(cfgfile, confpath);
   strcat(cfgfile, CONFIG_FILE);

   if ((cf = fopen(cfgfile, "r")) == NULL)
   {
#ifdef INIPATH
     strcpy(cfgfile, INIPATH);
     strcat(cfgfile, CONFIG_FILE);

     if ((cf = fopen(cfgfile, "r")) == NULL)
     {
#endif
#ifndef AXIP_CFG_BUGFIX
       xprintf("Config file %s not found or could not be opened\n", CONFIG_FILE);
#endif
       return(FALSE);
#ifdef INIPATH
     }
#endif
   }

   while (fgets(buf, 255, cf) != NULL)
   {
     (void)strcpy(cbuf, buf);
     ++lineno;

     if ((e = parse_line(buf)) < 0)
     {
#ifndef AXIP_CFG_BUGFIX
       xprintf("Config error at line %d: ", lineno);
#else
       xprintf("%s config error at line %d: ", CONFIG_FILE, lineno);
#endif

       switch (e)
       {
         case -1 : xprintf("Missing argument\n"); break;
         case -2 : xprintf("Bad callsign format\n"); break;
         case -3 : xprintf("Bad option - on/off\n"); break;
         case -4 : xprintf("Bad option - tnc/digi\n"); break;
         case -5 : xprintf("Host not known\n"); break;
         case -6 : xprintf("Unknown command\n"); break;
         case -7 : xprintf("Text string too long\n"); break;
         case -8 : xprintf("Bad option - every/after\n"); break;
         case -9 : xprintf("Bad option - ip/udp\n"); break;

         default: xprintf("Unknown error\n"); break;
       }
       xprintf("%s", cbuf);
#ifndef AXIP_CFG_BUGFIX
       ++errflag;
#endif
     }
   }

#ifndef AXIP_CFG_BUGFIX
   if (errflag)
     exit(1);
#endif

#ifndef AXIP_CFG_BUGFIX
   if (mode == NO_MODE)
   {
     xprintf("Must specify ip and/or udp sockets\n");
  #ifdef AXIP_CFG_BUGFIX_OLD
     fclose(cf);  /* Die Datei ist offen, also schliessen (db1ras) */
  #endif
     return(FALSE);
   }
#endif

   fclose(cf);

   return(TRUE);
}

/* Process each line from the config file.  The return value is encoded. */
int parse_line(char *buf)
{
   char *p;
   char *q;
   unsigned char tcall[7];
   unsigned char tip[4];
   struct hostent *he;
   int i = 0;
   int j = 0;
   int uport = 0;
   int dfalt = 0;

   p = strtok(buf, " \t\n\r");

   /* Leere Zeilen */
   if (p == NULL)
     return(0);

   /* Kommentarzeilen nicht auswerten */
   if (*p == '#')
     return(0);

   if (strcmp(p, "socket") == 0)
   {
     q = strtok(NULL, " \t\n\r");

     if (q == NULL)
       return(-1);

     if (strcmp(q, "ip") == 0)
       mode = IP_MODE;
     else if (strcmp(q, "udp") == 0)
     {
        mode = UDP_MODE;
        my_udp = htons(DEFAULT_UDP_PORT);

        /* Bei UDP optionale Portnummer lesen */
        q = strtok(NULL, " \t\n\r");

        /* Nummer war vorhanden, auswerten */
        if (q != NULL)
        {
           i = atoi(q);
           /* Nummer ist gueltig, uebernehmen */
           if (i > 0)
             my_udp = htons(i);
        }
     }
     else
       return(-9);

     return(0);
   }
   else if (strcmp(p, "socketoption") == 0)
     {
       q = strtok(NULL, " \t\n\r");

       /* Pruefung auf leere Zeile */
       if (q == NULL)
         return(-1);

       if (strcmp(q, "SO_KEEPALIVE") == 0)
       {
          sockopt_keepalive = TRUE;
       }
       else if (strcmp(q, "IPTOS_THROUGHPUT") == 0)
       {
          sockopt_throughput = TRUE;
       }
       else return(-9);

     return(0);

    } else if (strcmp(p, "loglevel") == 0) {
      q = strtok(NULL, " \t\n\r");
      if (q == NULL) return(-1);
      loglevel = atoi(q);
      return(0);

   } else if (strcmp(p, "route") == 0) {
      uport = 0;
      dfalt = 0;

      q = strtok(NULL, " \t\n\r");

      /* Pruefung auf leere Zeile */
      if (q == NULL)
        return(-1);

      /* Pruefung auf default-Route */
      if (strcmp(q, "default") == 0)
        dfalt = 1;
      else
      {
         /* Ziel der Route pruefen */
         if (a_to_call(q, tcall) != 0)
           return(-2);
      }

      q = strtok(NULL, " \t\n\r");
      if (q == NULL)
        return(-1);

      he = gethostbyname(q);
      if (he != NULL)
      {
        (void)memcpy(tip, he->h_addr_list[0], 4);
      }
      else
      {                  /* maybe user specified a numeric addr? */
         j = inet_addr(q);
         if (j == -1)
            return(-5);  /* if -1, bad deal!                    */

         (void)memcpy(tip, (char *)&j, 4);
      }

      q = strtok(NULL, " \t\n\r");

      if (q != NULL)
      {
        if (strcmp(q, "udp") == 0)
        {
          uport = DEFAULT_UDP_PORT;

          q = strtok(NULL, " \t\n\r");
          if (q != NULL)
          {
            i = atoi(q);
            if (i > 0)
               uport = i;
          }
        }
      }

      route_add(tip, tcall, uport, dfalt, 0);
      return(0);
   }
   return(-999);
}


/* Convert ascii callsign to internal format */
int a_to_call(char *text, unsigned char *tcall)
{
   register size_t  i = 0;
   int           ssid = 0;
   unsigned char c;

   if (strlen(text) == 0)
     return(-1);

   for (i = 0; i < 6; ++i)
     tcall[i] = (' ' << 1);

   tcall[6] = '\0';

   for (i = 0; i < strlen(text); ++i)
   {
     c = text[i];
     if (c == '-')
     {
       ssid = atoi(&text[i + 1]);
       if (ssid > 15)
         return(-1);
       tcall[6] = (ssid << 1);
       return(0);
     }

     if (islower(c))
       c = toupper(c);

     if (i > 5)
       return(-1);

     tcall[i] = (c << 1);
   }
   return(0);
}

/* Convert internal callsign to printable format */
char *call_to_a(unsigned char *tcall)
{
   int i = 0;
   int ssid = 0;
   char *tptr;
   static char t[10];

   for (i = 0, tptr = t; i < 6; ++i)
   {
     if (tcall[i] == (' ' << 1))
       break;
     *tptr = tcall[i] >> 1;
     tptr++;
   }

   ssid = (tcall[6] >> 1) & 0x0f;
   if (ssid > 0)
   {
     *tptr = '-';
     tptr++;

     if (ssid > 9)
     {
       *tptr = '1';
       tptr++;
       ssid -= 10;
     }
     *tptr = '0' + ssid;
     tptr++;
   }

   *tptr = '\0';
   return(&t[0]);
}

void write_log(const char *format, ...)
{
  FILE *fp;
  va_list arg_ptr;
  struct tm *lt;

  fp = fopen("ax25ip.log", "a+");

  if (fp != NULL)
  {
    lt = localtime(&sys_time);
    fprintf(fp, "%16.16s:", ctime(&sys_time));
    va_start(arg_ptr, format);
    vfprintf(fp, format, arg_ptr);
    va_end(arg_ptr);
    fprintf(fp, "\n");
    fclose(fp);
  }
}

#endif

/* End of os/linux/ax25ip.c */
